#include "../src/comdef.h"
#include "../src/classic.h"

/*
   address calculations:

   word_width = 0 => one 8-bit word per one address   
     xbus_width = 0 => 8-bit IF
       width = 0 => 8-bit data  => A = addr;      (uint8*) access;
     xbus_width = 1 => 16-bit IF
       width = 0 => 8-bit data  => A = addr << 1; (uint16*) ss.
       width = 1 => 16-bit data => A = addr;      (uint16*) access.
     xbus_width = 2 => 32-bit IF
       width = 0 => 8-bit data  => A = addr << 2; (uint32*) ss.
       width = 1 => 16-bit data => A = addr << 1; (uint32*) access.
       width = 2 => 32-bit data => A = addr;      (uint32*) access.

   word_width = 1 => one 16-bit word per one address   
     xbus_width = 1 => 16-bit IF
       width = 0 => 8-bit data  => A = addr;      (uint16*) ss.
       width = 1 => 16-bit data => A = addr >> 1; (uint16*) access.
     xbus_width = 2 => 32-bit IF
       width = 0 => 8-bit data  => A = addr << 1; (uint32*) access.
       width = 1 => 16-bit data => A = addr;      (uint32*) access.
       width = 2 => 32-bit data => A = addr >> 1; (uint32*) access.

   word_width = 2 => one 32-bit word per one address   
     xbus_width = 2 => 32-bit IF
       width = 0 => 8-bit data  => A = addr;      (uint32*) access.
       width = 1 => 16-bit data => A = addr >> 1; (uint32*) access.
       width = 2 => 32-bit data => A = addr >> 2; (uint32*) access.

*/

/*
  Write word of data to classic flash.
*/                
static void Classic_Write(uint32 addr, uint32 word)
{
  uint32 real_addr;
  uint32 real_word;
                     
  real_addr = (addr << (xbus_width-word_width-width)) & 
              (0xFFFFFFFFUL << (xbus_width - word_width));
  real_word = word << dbus_shift;

  switch (xbus_width) {
    case 0:
	  *(uint8*)(real_addr + base_address) = (uint8) real_word;
	  break;
	case 1:
	  *(uint16*)(real_addr + base_address) = (uint16) real_word;
	  break;
	case 2:
	  *(uint32*)(real_addr + base_address) = (uint32) real_word;
	  break;
  }

}      
                                      
/*
  Read word of data from classic flash.
*/                
uint32 Classic_Read(uint32 addr)
{

  uint32 real_addr;
  uint32 real_word;
  uint32 mask;

  real_addr = (addr << (xbus_width-word_width-width)) & 
              (0xFFFFFFFFUL << (xbus_width - word_width));

  switch (xbus_width) {
    case 0:
	  real_word = *(uint8*)(real_addr + base_address);
	  break;
	case 1:
	  real_word = *(uint16*)(real_addr + base_address);
	  break;
	case 2:
	  real_word = *(uint32*)(real_addr + base_address);
	  break;
  }

  mask = 0xFFUL;
  if (width == 1) mask = 0xFFFFUL;
  if (width == 2) mask = 0xFFFFFFFFUL;

  return (real_word >> dbus_shift) & mask;

}

/*
  Wait for classic flash operation complete.
*/                
static uint16 Classic_Wait(uint32 addr)
{
  uint16 data0, data1;
  
  data0 = Classic_Read(addr);
  do {
    data1 = Classic_Read(addr);
    if ( ((data1 ^ data0) & 0x40 ) == 0 )
      return 1; /* completed */      
    data0 = data1;  
  } while ( (data1 & 0x20) == 0);
  
  data0 = Classic_Read(addr);
  data1 = Classic_Read(addr);   
  
  return (((data1 ^ data0) & 0x40 ) == 0 ) ? 1 : 0;

}

static void Classic_Reset(void)
{
  Classic_Write(cmd_offsets[0], 0xAA);
  Classic_Write(cmd_offsets[1], 0x55);
  Classic_Write(cmd_offsets[0], 0xAA);
  Classic_Write(cmd_offsets[1], 0x55);
  Classic_Write(cmd_offsets[0], 0xF0);
}

uint16 perform_classic_flash_program()
{       

  uint32 word, data;
  uint16 i;          
  uint16 ofs = 0;  
  uint16 result;
  uint32 addr = address;
  uint16 incr;
  
  if (!check_iface_params()) 
    return 0;

  /* write "length" bytes from buffer */
  /* starting from "address" */
  
  incr = 1;
  if (width == 1) incr++;
  if (width == 2) incr += 2;

  /* Write data from buffer */
  for (i=0; i<length; i+=incr) {

    word = GetWord((void*)buffer, ofs);
	ofs += incr;
    
    /* Read data at current address, and compare with terget data */
	/* Program only if different */
    if (word != (data = Classic_Read(addr))) {

	  /* Check for programming can be completed: */
	  /* ones in the target data can be only on positions */
	  /* of ones in flash */
	  if (word & ~data) {
        /* Error - this data can't be programmed */
        status = STATUS_FLASH_DIRTY;      
        return 0;
	  }

      /* Programming sequence */  
      Classic_Write(cmd_offsets[0], 0xAA);
      Classic_Write(cmd_offsets[1], 0x55);
      Classic_Write(cmd_offsets[0], 0xA0);
      Classic_Write(addr, word);

	  /* Wait for programming completed */
      result = Classic_Wait(addr);
    
  	  /* Flash may report TIMEOUT! */
      if (!result) {
        Classic_Reset();
        status = STATUS_FLASH_TIMEOUT;      
        return 0;
      }          
	}
    
	/* Next address in flash */
    addr += incr;
  }

  /* Verify data written */

  for (i=0, ofs=0, addr = address; i<length; i+=incr) {

    word = GetWord((void*)buffer, ofs);
	ofs += incr;

	/* Compare with word from flash */
    if (word != Classic_Read(addr)) {

      status = STATUS_VERIFY_ERR;      
      return 0;

	}

	/* Next address in flash */
    addr += incr;
  }
 
  address = addr;

  return 1;
}
                         
uint16 perform_classic_flash_erase()
{     

  uint16 result;

  if (!check_iface_params()) 
    return 0;

  /* "address" points to any word within a sector to be erased */
  /* Sector erase sequence */
                    
  Classic_Write(cmd_offsets[0], 0xAA);
  Classic_Write(cmd_offsets[1], 0x55);
  Classic_Write(cmd_offsets[0], 0x80);
  Classic_Write(cmd_offsets[0], 0xAA);
  Classic_Write(cmd_offsets[1], 0x55);
  Classic_Write(address,  0x30);
 
  /* Wait for erasing completed */
  result = Classic_Wait(address);

  /* Flash may report TIMEOUT! */
  if (!result) {
    Classic_Reset();
    status = STATUS_FLASH_TIMEOUT;      
    return 0;
  }          
  
  return 1;
}

